#include "slurm_vipu.h"
#include "vipu_client.h"
#include "vipu_config.h"

/* Global variables and forward declarations */
typedef struct VipuClient VipuClient;
#define sv_debug(fmt, ...) debug("slurm/vipu: %s " fmt, __func__ , ##__VA_ARGS__)

#define VIPU_CONF_LEN 10
#define VIPU_CONF_SUFFIX ".vipu.conf"
#define MAX_PARTITION_NAME 64
#define MAX_IPUOF_CONFIG 32768
#define MAX_WAIT_PARTITION 800
#define MAX_RETRIES_REMOVE_PARTITION 10
#define DEFAULT_VIPU_MODEL_TYPE "default"

/* TODO: Refactor and rename structs */
typedef struct vipu_client {
	char *host;
	uint16_t port;
	char *user;
	char *accesskey;
	uint16_t secure;
	char *client_cert_file;
	char *client_key_file;
	char *server_cert_file;
	char *ipuof_dir;
	uint32_t timeout;
	uint32_t max_ipus_per_job;
	char *client_cert;
	char *client_key;
	char *server_cert;
	uint16_t force_delete_partition;
	uint16_t use_reconfig_partition;
	VipuClient *vipu_client;
} vipu_client_t;

typedef struct vipu_client_data {
	vipu_client_t v_client;
	vipu_client_t *v_client_ptr;
	bool vipu_client_initialized;
	vipu_config_data_t *vipu_conf_data;
} vipu_client_data_t;


extern char *vipu_p_partition_name(uint32_t job_id) {
	char *partition_name = NULL;

	xstrfmtcat(partition_name, "slurm_%u", job_id);
	return partition_name;
}

static inline char *vipu_p_get_vipu_config_file_name(const char *model) {
	char *file_name = NULL;

	xassert(model);
	xstrfmtcat(file_name, "%s"VIPU_CONF_SUFFIX, model);
	return file_name;
}

static inline char *vipu_p_get_vipu_model(const char *config_file)
{
	char *res = NULL;
	xassert(config_file);
	_xmemcat(&res, (char *)config_file, xstrstr(config_file, VIPU_CONF_SUFFIX));

	return res;
}

/*
 * validate_ipu_job - Validates the IPU job
 */
static int validate_ipu_job(unsigned count)
{
	if ((((count - 1) & count) == 0) && count < 128)
		return SLURM_SUCCESS;
	error("slurm_vipu: %u does not satisfy ipu count conditions(power of 2 and less than 128", count);
	return SLURM_ERROR;
}

/*
 * validate_ipu_job_for_client - Validates the IPU job for client
 */
static int validate_ipu_job_for_client(vipu_client_data_t *client, int num_ipus)
{
	sv_debug("entry");
	if (validate_ipu_job(num_ipus) != SLURM_SUCCESS)
		return SLURM_ERROR;

	xassert(client);

	if (num_ipus > client->v_client_ptr->max_ipus_per_job) {
		error("select_vipu invalid number of ipus requested: %d (max allowed: %d)",
		      num_ipus, client->v_client_ptr->max_ipus_per_job);
		return SLURM_ERROR;
	}

	return SLURM_SUCCESS;
}


/*
 * _read_file_to_str - Reads a file to a string
 */
static char *_read_file_to_str(const char *path)
{
	char *buffer = NULL;
	FILE *file = fopen(path, "r");
	long length;

	if (!file) {
		error("select_vipu cannot read file: %s", path);
		return NULL;
	}
	fseek(file, 0, SEEK_END);
	length = ftell(file);
	fseek(file, 0, SEEK_SET);
	buffer = malloc(sizeof(char) * (length + 1));
	buffer[length] = '\0';
	if (fread(buffer, sizeof(char), length, file) != length) {
		free(buffer);
		return NULL;
	}
	fclose(file);
	return buffer;
}

/*
 * vipu_client_destroy, destroys a vipu client
 */
static void vipu_p_client_destroy(vipu_client_data_t *conf)
{
	sv_debug(" destroying vipu client");
	if (!conf)
		return;

	vipu_conf_destroy(conf->vipu_conf_data);

	xfree(conf->v_client_ptr->host);
	xfree(conf->v_client_ptr->user);
	xfree(conf->v_client_ptr->accesskey);
	xfree(conf->v_client_ptr->client_cert_file);
	xfree(conf->v_client_ptr->client_key_file);
	xfree(conf->v_client_ptr->server_cert_file);
	xfree(conf->v_client_ptr->ipuof_dir);
	xfree(conf->v_client_ptr->client_cert);
	xfree(conf->v_client_ptr->client_key);
	xfree(conf->v_client_ptr->server_cert);
	conf->vipu_client_initialized = false;

	xfree(conf);
}

/*
 * vipu_client_create, creates a vipu client
 */
static vipu_client_data_t *vipu_p_client_create(const char *path)
{
	vipu_client_data_t *conf = NULL;

	if (!path)
		return NULL;

	conf = (vipu_client_data_t *)xcalloc(1, sizeof(vipu_client_data_t));
	if (!conf)
		return NULL;

	conf->v_client_ptr = &conf->v_client;
	conf->vipu_client_initialized = false;
	conf->vipu_conf_data = NULL;

	sv_debug("initializing vipu_client");

	/* Get the configuration */
	conf->vipu_conf_data = vipu_conf_init(path);
	if (!conf->vipu_conf_data) {
		sv_debug(" failed to initialize vipu configuration");
		xfree(conf);
		return NULL;
	}

	conf->v_client_ptr->host = xstrdup(vipu_conf_get_host(conf->vipu_conf_data));
	conf->v_client_ptr->port = vipu_conf_get_port(conf->vipu_conf_data);
	conf->v_client_ptr->user = xstrdup(vipu_conf_get_user(conf->vipu_conf_data));
	conf->v_client_ptr->accesskey =
	    xstrdup(vipu_conf_get_accesskey(conf->vipu_conf_data));
	conf->v_client_ptr->secure = vipu_conf_get_secure(conf->vipu_conf_data);
	conf->v_client_ptr->client_cert_file =
	    xstrdup(vipu_conf_get_client_cert_file(conf->vipu_conf_data));
	conf->v_client_ptr->client_key_file =
	    xstrdup(vipu_conf_get_client_key_file(conf->vipu_conf_data));
	conf->v_client_ptr->server_cert_file =
	    xstrdup(vipu_conf_get_server_cert_file(conf->vipu_conf_data));
	conf->v_client_ptr->ipuof_dir =
	    xstrdup(vipu_conf_get_ipuof_dir(conf->vipu_conf_data));
	conf->v_client_ptr->timeout = vipu_conf_get_timeout(conf->vipu_conf_data);
	conf->v_client_ptr->max_ipus_per_job =
	    vipu_conf_get_max_ipus_per_job(conf->vipu_conf_data);
	conf->v_client_ptr->force_delete_partition =
	    vipu_conf_get_force_delete_partition(conf->vipu_conf_data);
	conf->v_client_ptr->use_reconfig_partition =
	    vipu_conf_get_use_reconfig_partition(conf->vipu_conf_data);

	verbose("ApiHost                = %s", conf->v_client_ptr->host);
	verbose("ApiPort                = %u", conf->v_client_ptr->port);
	verbose("ApiUser                = %s", conf->v_client_ptr->user);
	verbose("IsSecure               = %u", conf->v_client_ptr->secure);
	verbose("ClientCertFile         = %s", conf->v_client_ptr->client_cert_file);
	verbose("ClientKeyFile          = %s", conf->v_client_ptr->client_key_file);
	verbose("ServerCertFile         = %s", conf->v_client_ptr->server_cert_file);
	verbose("IpuofDir               = %s", conf->v_client_ptr->ipuof_dir);
	verbose("ApiTimeout             = %d", conf->v_client_ptr->timeout);
	verbose("MaxIpusPerJob          = %d", conf->v_client_ptr->max_ipus_per_job);
	verbose("ForceDeletePartition   = %u",
		conf->v_client_ptr->force_delete_partition);
	verbose("UseReconfigPartition   = %u",
		conf->v_client_ptr->use_reconfig_partition);

	/* Host and port and required options for the ipu jobs */
	if (!conf->v_client_ptr->host || conf->v_client_ptr->port == 0) {
		error("select_vipu vipu host and port must be defined for the ipu jobs");
		vipu_p_client_destroy(conf);
		return NULL;
	}

	/* If secure mode is used then user, accesskey, and client/server certificates
	 * are needed */
	if (conf->v_client_ptr->secure) {
		if (!conf->v_client_ptr->user || !conf->v_client_ptr->accesskey ||
			!conf->v_client_ptr->client_cert_file ||
			!conf->v_client_ptr->client_key_file ||
			!conf->v_client_ptr->server_cert_file) {
			error("select_vipu vipu user id, access key, client cert/key, and server key "
			      "files must be defined for the ipu jobs on this cluster");
			vipu_p_client_destroy(conf);
			return NULL;
		}

		/* Read files */
		conf->v_client_ptr->client_cert =
		    _read_file_to_str(conf->v_client_ptr->client_cert_file);
		conf->v_client_ptr->client_key =
		    _read_file_to_str(conf->v_client_ptr->client_key_file);
		conf->v_client_ptr->server_cert =
		    _read_file_to_str(conf->v_client_ptr->server_cert_file);
		if (!conf->v_client_ptr->client_cert || !conf->v_client_ptr->client_key ||
			!conf->v_client_ptr->server_cert) {
			error("select_vipu could not read certificate files for the secure client");
			vipu_p_client_destroy(conf);
			return NULL;
		}
	}

	conf->vipu_client_initialized = true;
	return conf;
}

static bool vipu_p_client_init(vipu_client_data_t *client)
{
	xassert(client);
	xassert(client->v_client_ptr);

	sv_debug("trying vipu host address: %s", client->v_client_ptr->host);
	if (!client->v_client_ptr->host)
		return false;

	if (client->v_client_ptr->secure) {
		client->v_client_ptr->vipu_client = vipu_init_secure(
				client->v_client_ptr->host, client->v_client_ptr->port,
				client->v_client_ptr->timeout, client->v_client_ptr->client_cert,
				client->v_client_ptr->client_key, client->v_client_ptr->server_cert,
				client->v_client_ptr->user, client->v_client_ptr->accesskey);
	} else {
		client->v_client_ptr->vipu_client = vipu_init(
				client->v_client_ptr->host, client->v_client_ptr->port, client->v_client_ptr->timeout);
	}
	return client->v_client_ptr->vipu_client != NULL;
}

static void vipu_p_client_stop(vipu_client_data_t *client)
{
	xassert(client && client->v_client_ptr);
	vipu_destroy(client->v_client_ptr->vipu_client);
	client->v_client_ptr->vipu_client = NULL;
}

static vipu_client_data_t *vipu_p_client_create_from_conf(const char *config)
{
	xassert(config);
	const char *config_file = vipu_p_get_vipu_config_file_name(config);
	vipu_client_data_t *client;

	if (!config_file)
	return NULL;

	client = vipu_p_client_create(config_file);
	xfree(config_file);
	return client;
}

/*
 * list_vipu_config_files - list all vipu configuration files
 */
extern List vipu_p_get_configs(void)
{
	List config_files = NULL;
	char *dir_name = NULL;
	size_t lenstr;
	struct stat sbuf;
	DIR *f_dir;
	struct dirent *dir_ent;

	xstrfmtcat(dir_name, "%s", default_vipu_config);

	if (stat(dir_name, &sbuf)) {
		xfree(dir_name);
		return config_files;
	}

	f_dir = opendir(dir_name);
	if (f_dir) {
		config_files = list_create(xfree_ptr);
		while ((dir_ent = readdir(f_dir))) {
			if (!xstrcmp(dir_ent->d_name, ".") || !xstrcmp(dir_ent->d_name, ".."))
				continue;
			if (!list_find_first(config_files, slurm_find_char_in_list,
					     dir_ent->d_name)) {
				lenstr = strlen(dir_ent->d_name);
				if (lenstr > VIPU_CONF_LEN) {
					// TODO: Some sort of hardcoding file patterns
					if (xstrncmp(dir_ent->d_name + lenstr - VIPU_CONF_LEN,
							VIPU_CONF_SUFFIX, VIPU_CONF_LEN) == 0) {
						debug("slurm/vipu: config file found: %s", dir_ent->d_name);
						char *config = vipu_p_get_vipu_model(dir_ent->d_name);
						xassert(config);
						list_append(config_files, config);
					}
				}
			}
		}
		closedir(f_dir);
	} else {
		error("opendir(%s): %m", dir_name);
	}

	xfree(dir_name);
	return config_files;
}

extern int validate_ipu_job_for_config(const char *config, unsigned num_ipus)
{
	int max_ipus_per_job = 0;
	int retval = SLURM_ERROR;
	vipu_client_data_t *client = vipu_p_client_create_from_conf(config);

	if (!client) {
		debug("select_vipu client cannot be created for model %s", config);
		return retval;
	}
	retval = validate_ipu_job_for_client(client, num_ipus);
	vipu_p_client_destroy(client);

	return retval;
}

/*
 * create_partition - Creates a VIPU partition for the job using vipu_client
 */
static int create_partition(vipu_client_data_t *client, unsigned job_id, unsigned num_ipus, unsigned num_replicas, unsigned num_tasks)
{
	char *partition_name;
	char ipuof_config[MAX_IPUOF_CONFIG];
	int retval = SLURM_ERROR;

	partition_name = vipu_p_partition_name(job_id);
	if (!partition_name)
		return SLURM_ERROR;

	if (!vipu_create_partition_with_details(
					client->v_client_ptr->vipu_client,
					partition_name, num_ipus, "",
					0, num_tasks,
					client->v_client_ptr->use_reconfig_partition,
					num_replicas, ipuof_config)) {
		sv_debug("successful on vipu host address: %s",
			 client->v_client_ptr->host);
		retval = SLURM_SUCCESS;
	} else {
		sv_debug("could not create a partition(i:%u,r:%u) for the job:%u at this time\
		  , trying other vipu models and/or queued for later",
			 num_ipus, num_replicas, job_id);
	}
	xfree(partition_name);

	return retval;
}

/*
 * create_partition_for_model() - Create a partition by creating client for the
 * model
 */
extern int vipu_p_create_partition(const char *config, unsigned num_ipus, unsigned num_replicas, unsigned job_id, unsigned num_tasks)
{
	vipu_client_data_t *client = vipu_p_client_create_from_conf(config);
	int retval = SLURM_SUCCESS;

	sv_debug("called for the job %u with model %s", job_id, config);

	if (!client) {
		error("select_vipu client cannot be created for model %s", config);
		return SLURM_ERROR;
	}

	if (validate_ipu_job_for_client(client, num_ipus) != SLURM_SUCCESS) {
		vipu_p_client_destroy(client);
		return SLURM_ERROR;
	}

	if (!vipu_p_client_init(client)) {
		vipu_p_client_destroy(client);
		error("select_vipu: Failed to initialize VIPU client");
		return SLURM_ERROR;
	}

	retval = create_partition(client, job_id, num_ipus, num_replicas, num_tasks);
	if (retval != SLURM_SUCCESS) {
		error("select_vipu creating a partition for the job %u with config %s failed",
		      job_id, config);
	}

	vipu_p_client_stop(client);
	vipu_p_client_destroy(client);
	sv_debug("%s: exit creating a partition for the job %u with model %s",
	      __func__, job_id, config);
	return retval;
}

/*
 * remove_partition - Removes a VIPU partition for the job using vipu_client
 */
static int remove_partition(vipu_client_data_t *client, const char *partition_name)
{
	int i;

	xassert(client);

	if (!partition_name)
		return SLURM_ERROR;

	sv_debug("partition '%s' is on host %s", partition_name,
		 client->v_client_ptr->host);

	for (i = 0; i < MAX_RETRIES_REMOVE_PARTITION; i++) {
		int retval;

		if (!client->v_client_ptr->force_delete_partition) {
			retval = vipu_remove_partition(client->v_client_ptr->vipu_client,
						       partition_name);
		} else {
			retval = vipu_remove_partition_force(client->v_client_ptr->vipu_client,
							     partition_name);
		}
		if (retval != 0)
			//TODO: Sleep
			sleep(1);
		else
			return SLURM_SUCCESS;
	}
	return SLURM_ERROR;
}

extern int vipu_p_destroy_partition(const char *partition_name, const char *config)
{
	xassert(partition_name && config);
	vipu_client_data_t * client;
	int i, retval = SLURM_ERROR;
	unsigned num_gcds;
	char **gcd_configs;

	client = vipu_p_client_create_from_conf(config);
	if (!client)
		return SLURM_ERROR;
	if(!vipu_p_client_init(client)) {
		vipu_p_client_destroy(client);
 		error("select_vipu: Failed to initialize VIPU client");
		return SLURM_ERROR;
	}

	if (remove_partition(client, partition_name) == SLURM_ERROR) {
		error("Could not remove partition(%s)", partition_name);
		goto release_client;
	}

	gcd_configs = vipu_get_ipuof_config(client->v_client_ptr->vipu_client,
					    partition_name, &num_gcds);
	if (!gcd_configs) {
		error("select_vipu could not get ipuof config of the partition %s", partition_name);
		goto release_client;
	}

	for (i = 0; i < num_gcds; i++) {
		char temp_file[256];

		snprintf(temp_file, 256, "%s/%s_%d_ipuof.conf",
			client->v_client_ptr->ipuof_dir, partition_name, i);
		temp_file[255] = 0;
		if (remove(temp_file) != 0) {
			error("select_vipu could not remove IPUoF config file for the '%s' - errno: %d",
			      temp_file, errno);
		}
	}

	retval = SLURM_SUCCESS;
	for (i = 0; i < num_gcds; i++)
		free(gcd_configs[i]);

	if (gcd_configs)
		free(gcd_configs);
release_client:
	vipu_p_client_stop(client);
	vipu_p_client_destroy(client);

	return retval;
}

extern int vipu_p_is_partition_active(const char *partition_name, const char *config) {
	xassert(partition_name && config);
	partition_state_t status;
	vipu_client_data_t *client = vipu_p_client_create_from_conf(config);

	if (!client)
		return SLURM_ERROR;

	if (!vipu_p_client_init(client)) {
		vipu_p_client_destroy(client);
		error("Failed to initialize VIPU client");
		return SLURM_ERROR;
	}
	if (vipu_get_partition_status(client->v_client_ptr->vipu_client,
				  partition_name, &status)) {
	error("Failed to get %s partition status", partition_name);
	return SLURM_ERROR;
    }

	return status == PS_ACTIVE ? SLURM_SUCCESS : SLURM_ERROR;
}


/*
 * copy_ipuof_config - Gets a copies IPUoF configuration file for a job
 */
static int copy_ipuof_config(vipu_client_data_t *client, const char *partition_name)
{
	char temp_file[PATH_MAX];
	char **gcd_configs;
	FILE *fp;
	uint32_t i, num_gcds = 0;

	xassert(client);

	gcd_configs = vipu_get_ipuof_config(client->v_client_ptr->vipu_client,
					    partition_name, &num_gcds);
	if (!gcd_configs) {
		sv_debug(" could not get ipuof config of the partition %s at this time, queued for later",
			 partition_name);

		return SLURM_ERROR;
	}

	debug("Total gcd configuration for this partition are: %d", num_gcds);
	for (i = 0; i < num_gcds; i++) {
		char *file_str = xstrdup(gcd_configs[i]);

		sv_debug(" configuration for gcd %d is %s", i, gcd_configs[i]);
		sprintf(temp_file, "%s/%s_%d_ipuof.conf",
			client->v_client_ptr->ipuof_dir, partition_name, i);

		sv_debug(" temp file for gcd %d is %s", i, temp_file);
		sv_debug(" file configuration is %s", file_str);
		fp = fopen(temp_file, "w");
		if (fp) {
			if (fputs(gcd_configs[i], fp) == EOF) {
				error("select_vipu error writing IPUoF configuration file for %s",
				      partition_name);
				fclose(fp);
			} else {
				sv_debug(" IPUoF configuration for the %s is written",
					 partition_name);
			    fclose(fp);
			}
		} else {
			error("select_vipu could not open file for writing - errno: %d", errno);
		}
		xfree(file_str);
	}

	for (i = 0; i < num_gcds; i++)
		free(gcd_configs[i]);
	if (gcd_configs)
		free(gcd_configs);

	return SLURM_SUCCESS;
}

extern int vipu_p_copy_ipuof_configs(const char *partition_name, const char *config)
{
	xassert(partition_name && config);
	vipu_client_data_t *client = vipu_p_client_create_from_conf(config);
	int retval;

	if (!client)
		return SLURM_ERROR;

	if (!vipu_p_client_init(client)) {
		error("select_vipu cannot create vipu client for job info");
		return SLURM_ERROR;
	}

	retval = copy_ipuof_config(client, partition_name);
	vipu_p_client_stop(client);
	vipu_p_client_destroy(client);
	return retval;
}
