/*****************************************************************************\
 * Copyright (c) 2021-2022 Graphcore Ltd.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 *
 * This file is part of the modified version of Slurm, a resource management
 * program with Graphcore IPU support. The modified version includes a new
 * resource selection plugin, vipu, as well as a new job submission plugin,
 * vipu, with the support handling IPU jobs.
 *
 * Please also read the included files: README, DISCLAIMER and COPYING.
 *
 * The modified Slurm source code is modified and redistributed under the terms
 * of the GNU General Public License version 3 (GPL v3) as published by the
 * Free Software Foundation.
 *
\*****************************************************************************/

/*****************************************************************************\
 * config.c
 * Configuration handling for the Resource selection plugin with VIPU support
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "src/common/slurm_xlator.h"
#include "slurm/slurm.h"
#include "src/common/fetch_config.h"
#include "src/common/list.h"
#include "src/common/log.h"
#include "src/common/macros.h"
#include "src/common/parse_config.h"
#include "src/common/strlcpy.h"
#include "src/common/xmalloc.h"
#include "src/common/xstring.h"

#include "vipu_config.h"

s_p_options_t vipu_conf_options[] = {
    {"ApiHost", S_P_STRING},
    {"ApiPort", S_P_UINT16},
    {"IsSecure", S_P_UINT16},
    {"ApiUser", S_P_STRING},
    {"ApiAccessKey", S_P_STRING},
    {"ClientCertFile", S_P_STRING},
    {"ClientKeyFile", S_P_STRING},
    {"ServerCertFile", S_P_STRING},
    {"IpuofDir", S_P_STRING},
    {"MaxIpusPerJob", S_P_UINT32},
    {"ApiTimeout", S_P_UINT32},
    {"ForceDeletePartition", S_P_UINT16},
    {"UseReconfigPartition", S_P_UINT16},
    {NULL}
};

/*
 * vipu_conf_init - load the vipu configuration from a file.
 */
extern vipu_config_data_t *vipu_conf_init(const char *path)
{
    int rc;
    char *config_file = NULL;

    vipu_config_data_t *conf = NULL;

    conf = (vipu_config_data_t *)xmalloc(
        sizeof(vipu_config_data_t));

    conf->vipu_conf = (vipu_select_config_t){
        NULL, /* model */
        NULL, /* host */
        0, /* port */
        0, /* secure */
        NULL, /* user */
        NULL, /* accesskey */
        NULL, /* client_cert_file */
        NULL, /* client_key_file */
        NULL, /* server_cert_file */
        NULL, /* ipuof_dir */
        0,    /* max_ipus_per_job */
        0,    /* timeout */
        0,    /* hash_val */
        0,    /* force_delete_partition */
        0     /* use_reconfig_partition */
    };

    conf->vipu_conf_hashtbl = NULL;
    conf->vipu_conf_ptr = &conf->vipu_conf;
    conf->vipu_conf_initialized = false;

    if (path == NULL)
        xstrfmtcat(config_file, "%s%s", default_vipu_config, "default.vipu.conf");
    else
        xstrfmtcat(config_file, "%s%s", default_vipu_config, path);

    debug("vipu: config: configuration file is %s", config_file);

    xfree(conf->vipu_conf_ptr->model);
    xfree(conf->vipu_conf_ptr->host);
    xfree(conf->vipu_conf_ptr->user);
    xfree(conf->vipu_conf_ptr->accesskey);
    xfree(conf->vipu_conf_ptr->client_cert_file);
    xfree(conf->vipu_conf_ptr->client_key_file);
    xfree(conf->vipu_conf_ptr->server_cert_file);
    xfree(conf->vipu_conf_ptr->ipuof_dir);

    if (access(config_file, F_OK) != 0)
    {
        // file does not exist
        error("vipu: config: configuration file %s not found", config_file);
        xfree(config_file);
        xfree(conf);
        return NULL;
    }

    conf->vipu_conf_hashtbl = s_p_hashtbl_create(vipu_conf_options);

    conf->vipu_conf_ptr->hash_val = 0;
    rc = s_p_parse_file(conf->vipu_conf_hashtbl, &conf->vipu_conf_ptr->hash_val, config_file, false, NULL);

    if (rc != SLURM_SUCCESS)
    {
        error("vipu: config: error while parsing file %s", config_file);
        s_p_hashtbl_destroy(conf->vipu_conf_hashtbl);
        xfree(config_file);
        xfree(conf);
        return NULL;
    }

    if (!s_p_get_string(&conf->vipu_conf_ptr->host, "ApiHost", conf->vipu_conf_hashtbl))
    {
        error("vipu: config: missing value for 'ApiHost=' in configuration file %s", config_file);
        s_p_hashtbl_destroy(conf->vipu_conf_hashtbl);
        xfree(config_file);
        xfree(conf);
        return NULL;
    }

    if (!s_p_get_uint16(&conf->vipu_conf_ptr->port, "ApiPort", conf->vipu_conf_hashtbl))
        conf->vipu_conf_ptr->port = DEFAULT_PORT;

    if (!s_p_get_uint16(&conf->vipu_conf_ptr->secure, "IsSecure", conf->vipu_conf_hashtbl))
        conf->vipu_conf_ptr->port = DEFAULT_SECURE;

    (void)s_p_get_string(&conf->vipu_conf_ptr->user, "ApiUser",
                         conf->vipu_conf_hashtbl);

    (void)s_p_get_string(&conf->vipu_conf_ptr->accesskey, "ApiAccessKey",
                         conf->vipu_conf_hashtbl);

    if (!s_p_get_string(&conf->vipu_conf_ptr->client_cert_file, "ClientCertFile", conf->vipu_conf_hashtbl))
        conf->vipu_conf_ptr->client_cert_file = xstrdup(DEFAULT_CLIENT_CERT_FILE);

    if (!s_p_get_string(&conf->vipu_conf_ptr->client_key_file, "ClientKeyFile", conf->vipu_conf_hashtbl))
        conf->vipu_conf_ptr->client_key_file = xstrdup(DEFAULT_CLIENT_KEY_FILE);

    if (!s_p_get_string(&conf->vipu_conf_ptr->server_cert_file, "ServerCertFile", conf->vipu_conf_hashtbl))
        conf->vipu_conf_ptr->server_cert_file = xstrdup(DEFAULT_SERVER_CERT_FILE);

    if (!s_p_get_string(&conf->vipu_conf_ptr->ipuof_dir, "IpuofDir", conf->vipu_conf_hashtbl))
        conf->vipu_conf_ptr->ipuof_dir = xstrdup(_DEFAULT_IPUOF_DIR);

    if (!s_p_get_uint32(&conf->vipu_conf_ptr->max_ipus_per_job, "MaxIpusPerJob", conf->vipu_conf_hashtbl))
        conf->vipu_conf_ptr->max_ipus_per_job = DEFAULT_MAX_IPUS_PER_JOB;

    if (!s_p_get_uint32(&conf->vipu_conf_ptr->timeout, "ApiTimeout", conf->vipu_conf_hashtbl))
        conf->vipu_conf_ptr->timeout = DEFAULT_TIMEOUT;

    if (!s_p_get_uint16(&conf->vipu_conf_ptr->force_delete_partition, "ForceDeletePartition", conf->vipu_conf_hashtbl))
        conf->vipu_conf_ptr->force_delete_partition = DEFAULT_FORCE_DELETE_PARTITION;

    if (!s_p_get_uint16(&conf->vipu_conf_ptr->use_reconfig_partition, "UseReconfigPartition", conf->vipu_conf_hashtbl))
        conf->vipu_conf_ptr->use_reconfig_partition = DEFAULT_USE_RECONFIG_PARTITION;

    conf->vipu_conf_ptr->model = xstrdup(path);
    conf->vipu_conf_initialized = true;

    debug("vipu: conf: successfully parsed configuration file");

    return conf;
}

/*
 * vipu_conf_destroy - Frees allocated memory for the vipu configuration
 */
extern int vipu_conf_destroy(vipu_config_data_t *conf)
{

    if (conf != NULL && conf->vipu_conf_initialized)
    {

        s_p_hashtbl_destroy(conf->vipu_conf_hashtbl);
        xfree(conf->vipu_conf_ptr->host);
        xfree(conf->vipu_conf_ptr->user);
        xfree(conf->vipu_conf_ptr->accesskey);
        xfree(conf->vipu_conf_ptr->client_cert_file);
        xfree(conf->vipu_conf_ptr->client_key_file);
        xfree(conf->vipu_conf_ptr->server_cert_file);
        xfree(conf->vipu_conf_ptr->ipuof_dir);
        xfree(conf->vipu_conf_ptr->model);
        conf->vipu_conf_initialized = false;

        xfree(conf);
    }

    return 0;
}

/*
 * vipu_conf_get_model
 */
extern char *vipu_conf_get_model(vipu_config_data_t *conf)
{
    char *model = NULL;
    if (!conf->vipu_conf_initialized)
        return NULL;
    model = xstrdup(conf->vipu_conf_ptr->model);
    return model;
}

/*
 * vipu_conf_get_host
 */
extern char *vipu_conf_get_host(vipu_config_data_t *conf)
{
    char *host = NULL;
    if (!conf->vipu_conf_initialized)
        return NULL;
    host = xstrdup(conf->vipu_conf_ptr->host);
    return host;
}

/*
 * vipu_conf_get_user
 */
extern char *vipu_conf_get_user(vipu_config_data_t *conf)
{
    char *user = NULL;
    if (!conf->vipu_conf_initialized)
        return NULL;
    user = xstrdup(conf->vipu_conf_ptr->user);
    return user;
}

/*
 * vipu_conf_get_accesskey
 */
char *vipu_conf_get_accesskey(vipu_config_data_t *conf)
{
    char *accesskey = NULL;
    if (!conf->vipu_conf_initialized)
        return NULL;
    accesskey = xstrdup(conf->vipu_conf_ptr->accesskey);
    return accesskey;
}

/*
 * vipu_conf_get_client_cert_file
 */
char *vipu_conf_get_client_cert_file(vipu_config_data_t *conf)
{
    char *client_cert_file = NULL;
    if (!conf->vipu_conf_initialized)
        return NULL;
    client_cert_file = xstrdup(conf->vipu_conf_ptr->client_cert_file);
    return client_cert_file;
}

/*
 * vipu_conf_get_client_key_file
 */
char *vipu_conf_get_client_key_file(vipu_config_data_t *conf)
{
    char *client_key_file = NULL;
    if (!conf->vipu_conf_initialized)
        return NULL;
    client_key_file = xstrdup(conf->vipu_conf_ptr->client_key_file);
    return client_key_file;
}

/*
 * vipu_conf_get_server_cert_file
 */
char *vipu_conf_get_server_cert_file(vipu_config_data_t *conf)
{
    char *server_cert_file = NULL;
    if (!conf->vipu_conf_initialized)
        return NULL;
    server_cert_file = xstrdup(conf->vipu_conf_ptr->server_cert_file);
    return server_cert_file;
}

/*
 * vipu_conf_get_ipuof_dir
 */
char *vipu_conf_get_ipuof_dir(vipu_config_data_t *conf)
{
    char *ipuof_dir = NULL;
    if (!conf->vipu_conf_initialized)
        return NULL;
    ipuof_dir = xstrdup(conf->vipu_conf_ptr->ipuof_dir);
    return ipuof_dir;
}

/*
 * vipu_conf_get_port
 */
uint16_t vipu_conf_get_port(vipu_config_data_t *conf)
{
    if (!conf->vipu_conf_initialized)
        return 0;
    return conf->vipu_conf_ptr->port;
}

/*
 * vipu_conf_get_secure
 */
uint16_t vipu_conf_get_secure(vipu_config_data_t *conf)
{
    if (!conf->vipu_conf_initialized)
        return 0;
    return conf->vipu_conf_ptr->secure;
}

/*
 * vipu_conf_get_max_ipus_per_job
 */
uint32_t vipu_conf_get_max_ipus_per_job(vipu_config_data_t *conf)
{
    if (!conf->vipu_conf_initialized)
        return 0;
    return conf->vipu_conf_ptr->max_ipus_per_job;
}

/*
 * vipu_conf_get_timeout
 */
uint32_t vipu_conf_get_timeout(vipu_config_data_t *conf)
{
    if (!conf->vipu_conf_initialized)
        return 0;
    return conf->vipu_conf_ptr->timeout;
}

/*
 * vipu_conf_get_force_delete_partition
 */
uint16_t vipu_conf_get_force_delete_partition(vipu_config_data_t *conf)
{
    if (!conf->vipu_conf_initialized)
        return 0;
    return conf->vipu_conf_ptr->force_delete_partition;
}

/*
 * vipu_conf_get_use_reconfig_partition
 */
uint16_t vipu_conf_get_use_reconfig_partition(vipu_config_data_t *conf)
{
    if (!conf->vipu_conf_initialized)
        return 0;
    return conf->vipu_conf_ptr->use_reconfig_partition;
}
