/*****************************************************************************\
 * Copyright (c) 2021-2022 Graphcore Ltd.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 *
 * This file is part of the modified version of Slurm, a resource management
 * program with Graphcore IPU support. The modified version includes a new
 * resource selection plugin, vipu, as well as a new job submission plugin,
 * vipu, with the support handling IPU jobs.
 *
 * Please also read the included files: README, DISCLAIMER and COPYING.
 *
 * The modified Slurm source code is modified and redistributed under the terms
 * of the GNU General Public License version 3 (GPL v3) as published by the
 * Free Software Foundation.
 *
 \*****************************************************************************/

/*****************************************************************************\
 * select_vipu.c - Resource selection plugin with VIPU support.
 * This plugin uses Slurm's 'other' select plugin interface to allow for using
 * it in conjunction with linear, cons_res, and cons_tres select plugins.
 *****************************************************************************/

#include <assert.h>
#include <fcntl.h>
#include <poll.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "src/common/slurm_xlator.h"
#include "slurm/slurm.h"
#include "config.h"
#include "other_select.h"
#include "src/bcast/file_bcast.h"
#include "src/common/gres.h"
#include "src/common/macros.h"
#include "src/common/pack.h"
#include "src/common/slurm_accounting_storage.h"
#include "src/slurmctld/burst_buffer.h"
#include "src/slurmctld/locks.h"

#include "slurm_vipu.h"

#define sv_debug(fmt, ...) debug("select/vipu: %s " fmt, __func__ , ##__VA_ARGS__)

enum VIPU_PARTITION_STATE
{
	INIT,
	VALIDATED,
	PARTITION_CREATED
};

/**
 * struct select_jobinfo - data specific to vipu node selection plugin
 * @magic:		magic number, must equal %JOBINFO_MAGIC
 * @other_jobinfo:	hook into attached, "other" node selection plugin.
 */
struct select_jobinfo {
	uint16_t magic;
	select_jobinfo_t *other_jobinfo;
	uint32_t state;
	uint64_t ipus;
	uint64_t num_replicas;
	char *config;
};

#define JOBINFO_MAGIC 0x86fa

/**
 * struct select_nodeinfo - data used for node information
 * @magic:		magic number, must equal %NODEINFO_MAGIC
 * @other_nodeinfo:	hook into attached, "other" node selection plugin.
 */
struct select_nodeinfo {
	uint16_t magic;
	select_nodeinfo_t *other_nodeinfo;
};

#define NODEINFO_MAGIC 0x88fa

// Slurm globals
int node_record_count;
slurm_conf_t slurmctl_config;
void *acct_db_conn;

static pthread_mutex_t vipu_config_list_lock = PTHREAD_MUTEX_INITIALIZER;
static List vipu_config_list;

/*
 * These variables are required by the generic plugin interface.  If they
 * are not found in the plugin, the plugin loader will ignore it.
 *
 * plugin_name - a string giving a human-readable description of the
 * plugin.  There is no maximum length, but the symbol must refer to
 * a valid string.
 *
 * plugin_type - a string suggesting the type of the plugin or its
 * applicability to a particular form of data or method of data handling.
 * If the low-level plugin API is used, the contents of this string are
 * unimportant and may be anything.  Slurm uses the higher-level plugin
 * interface which requires this string to be of the form
 *
 *	<application>/<method>
 *
 * where <application> is a description of the intended application of
 * the plugin (e.g., "select" for Slurm node selection) and <method>
 * is a description of how this plugin satisfies that application.  Slurm will
 * only load select plugins if the plugin_type string has a
 * prefix of "select/".
 *
 * plugin_version - an unsigned 32-bit integer containing the Slurm version
 * (major.minor.micro combined into a single number).
 */

const char plugin_name[] = "select/vipu: V-IPU Resource selection plugin";
const char plugin_type[] = "select/vipu";
const char vipu_gres_name[] = "vipu";

const uint32_t plugin_version = SLURM_VERSION_NUMBER;
uint32_t plugin_id = SELECT_PLUGIN_VIPU_LINEAR;

static inline select_jobinfo_t *jobinfo_from_job_record(job_record_t *job_ptr)
{
	select_jobinfo_t *jobinfo;

	xassert(job_ptr);
	xassert(job_ptr->select_jobinfo);
	xassert(job_ptr->select_jobinfo->data);

	jobinfo = job_ptr->select_jobinfo->data;

	xassert(jobinfo->magic == JOBINFO_MAGIC);

	return jobinfo;
}

static inline void print_gres(List gres_list)
{
	gres_state_t *job_gres_ptr;
	ListIterator job_gres_iter;

	if (!gres_list)
		return;

	job_gres_iter = list_iterator_create(gres_list);

	while ((job_gres_ptr = (gres_state_t *)list_next(job_gres_iter))) {
		if (job_gres_ptr->state_type != GRES_STATE_TYPE_JOB) {
			continue;
		}
		gres_job_state_t *job_gres_state = (gres_job_state_t *)
							 job_gres_ptr->gres_data;
		sv_debug("JOB GRES: %s:%s:%lu", job_gres_ptr->gres_name,
			 job_gres_state->type_name, job_gres_state->total_gres);
	}

	list_iterator_destroy(job_gres_iter);
}

static inline uint64_t get_count_for_gres(List gres_list, const char *query)
{
	uint64_t count = 0;
	gres_state_t *job_gres_ptr;
	ListIterator job_gres_iter;

	if (!gres_list)
		return 0;

	job_gres_iter = list_iterator_create(gres_list);

	while ((job_gres_ptr = (gres_state_t *)list_next(job_gres_iter))) {
		if (job_gres_ptr->state_type != GRES_STATE_TYPE_JOB)
			continue;
		if (xstrcmp(job_gres_ptr->gres_name, query)) {
			sv_debug("gres_name is %s, expected %s", job_gres_ptr->gres_name, query);
			continue;
		}
		gres_job_state_t *job_gres_state = (gres_job_state_t *)
							 job_gres_ptr->gres_data;
		sv_debug("found GRES = %s with type = %s count = %lu",
			job_gres_ptr->gres_name,
			job_gres_state->type_name,
			job_gres_state->total_gres);
		count = job_gres_state->total_gres;
	}

	list_iterator_destroy(job_gres_iter);
	return count;
}

/*
 * _get_job_vipu_gres_model - Get vipu GRES model for the job
 */
static char *get_job_vipu_gres_model(List job_gres_list)
{
	char *vipu_gres_model = NULL;

	if (gres_init() == SLURM_SUCCESS) {
		if (job_gres_list) {
			vipu_gres_model = gres_get_model_by_type(job_gres_list, (char *)vipu_gres_name);
			sv_debug(" gres model: %s", vipu_gres_model);
		} else {
			error("select_vipu failed to fetch GRES list for the job, using default default.vipu.conf");
		}
	} else {
		error("select_vipu failed to initialize gres plugin, using default.vipu.conf");
		return NULL;
	}
	return vipu_gres_model;
}

/*
 * _set_job_vipu_gres_model - Set vipu GRES model for the job
 */
static int set_job_vipu_gres_model(List job_gres_list, char *model)
{
	int ret = SLURM_ERROR;

	if (gres_init() == SLURM_SUCCESS) {
		if (job_gres_list) {
			ret = gres_set_model_by_type(job_gres_list, (char*)vipu_gres_name, model);
		} else {
			error("select_vipu failed to fetch GRES list for the job");
		}
	} else {
		error("select_vipu failed to initialize gres plugin");
	}
	return ret;
}

/*
 * release_job_resources - Releases any job resources allocated in case of job
 * termination or cancellation
 */
static int release_job_resources(job_record_t *job_ptr)
{
	select_jobinfo_t *jobinfo = jobinfo_from_job_record(job_ptr);
	uint64_t num_ipus = jobinfo->ipus;
	char *partition_name;
	char *model;

	sv_debug("Entry");
	/*
	 * Release job if there are no IPUs or job has not started
	 */
	if (!num_ipus || jobinfo->state != PARTITION_CREATED)
		return SLURM_SUCCESS;

	model = jobinfo->config;
	if (!model)
		return SLURM_ERROR;

	info("Relasing IPU resources for job = %u", job_ptr->job_id);
	partition_name = vipu_p_partition_name(job_ptr->job_id);
	if (!partition_name)
		return SLURM_ERROR;
	sv_debug("partition name = %s, model = %s", partition_name, model);

	vipu_p_destroy_partition(partition_name, model);

	xfree(partition_name);
	return SLURM_SUCCESS;
}

static int request_job_resources(job_record_t *job_ptr) {
	select_jobinfo_t *jobinfo = jobinfo_from_job_record(job_ptr);
	unsigned num_ipus = jobinfo->ipus;
	unsigned num_replicas = jobinfo->num_replicas;
	unsigned job_id = job_ptr->job_id;
	unsigned num_tasks = job_ptr->details->num_tasks;
	const char *config = jobinfo->config;

	info("Requesting IPU resources for job = %u", job_ptr->job_id);

	if (vipu_p_create_partition(config, num_ipus, num_replicas, job_id, num_tasks)) {
		error("Failed to create partition for job %u %s, %u:%u:%u",
			  job_id, config, num_ipus, num_replicas, num_tasks);
		return SLURM_ERROR;
	}
	return SLURM_SUCCESS;
}

/*
 * is_partition_active - Checks if the partition for the job ACTIVE
 */
static int is_partition_active(job_record_t *job_ptr)
{
	select_jobinfo_t *jobinfo = jobinfo_from_job_record(job_ptr);
	char *partition_name = vipu_p_partition_name(job_ptr->job_id);

	sv_debug("checking partition for the job: %u",
		 job_ptr->job_id);

	if (jobinfo->state != PARTITION_CREATED) {
		error("select_vipu partition does not exists in job details for %u",
			  job_ptr->job_id);
		return SLURM_ERROR;
	}
	sv_debug("partition is on %s for the job %u",
		 jobinfo->config, job_ptr->job_id);

	if (vipu_p_is_partition_active(partition_name, jobinfo->config))
		return SLURM_ERROR;

	return SLURM_SUCCESS;
}

/*
 * copy_ipuof_config - Gets a copies IPUoF configuration file for a job
 */
static int copy_ipuof_config(job_record_t *job_ptr)
{
	select_jobinfo_t *jobinfo = jobinfo_from_job_record(job_ptr);
	char *partition_name = vipu_p_partition_name(job_ptr->job_id);
	sv_debug(" copy_ipuof_config called");
	xassert(client);

	if (!partition_name)
		return SLURM_ERROR;

	if (vipu_p_copy_ipuof_configs(partition_name, jobinfo->config)) {
		error("Failed to copy IPUoF configuration for %s", partition_name);
		return SLURM_ERROR;
	}

	return SLURM_SUCCESS;
}

/*
 * Determine which of these nodes are usable by this job based on set job gres
 * for the vipu
 *
 * Remove nodes from node_bitmap that don't have current vipu GRES set
 *
 * Return SLURM_ERROR if a required node can't be used.
 *
 * IN job_ptr - pointer to job being considered for initiation,
 *                  set's start_time when job expected to start
 * IN/OUT bitmap - usable nodes are set on input, nodes not required to
 *	satisfy the request are cleared, other left set
 * IN min_nodes - minimum count of nodes
 * IN req_nodes - requested (or desired) count of nodes
 * IN max_nodes - maximum count of nodes
 * IN mode - SELECT_MODE_RUN_NOW: try to schedule job now
 *           SELECT_MODE_TEST_ONLY: test if job can ever run
 *           SELECT_MODE_WILL_RUN: determine when and where job can run
 * IN preemptee_candidates - List of pointers to jobs which can be preempted.
 * IN preemptee_job_list - Pointer to list of job pointers. These are the
 *		jobs to be preempted to initiate the pending job. Not set
 *		if mode=SELECT_MODE_TEST_ONLY or input pointer is NULL.
 * IN exc_cores - bitmap of cores being reserved.
 */
static int _filter_nodes_gres(job_record_t *job_ptr, bitstr_t *node_bitmap,
				  uint32_t min_nodes, uint32_t max_nodes,
				  uint32_t req_nodes, uint16_t mode,
				  List preemptee_candidates,
				  List *preemptee_job_list, bitstr_t *exc_cores)
{
	List gres_list;
	int i, i_first, i_last;

	i_first = bit_ffs(node_bitmap);
	if (i_first == -1)
		i_last = -2;
	else
		i_last = bit_fls(node_bitmap);

	for (i = i_first; i <= i_last; i++) {
		if (!bit_test(node_bitmap, i))
			continue;

		char *s = bit_fmt_full(node_bitmap);

		gres_list = node_record_table_ptr[i]->gres_list;
		sv_debug("Checking node %s node_bitmap = %s", node_record_table_ptr[i]->name, s);
		xfree(s);

		if (gres_plugin_job_test_model(job_ptr->gres_list_req,
						   gres_list) != SLURM_SUCCESS) {
			sv_debug(" Node %s is not usable for job %u due to GRES mismatch",
				  node_record_table_ptr[i]->name, job_ptr->job_id);
			bit_clear(node_bitmap, i);
			if (job_ptr->details->req_node_bitmap &&
				bit_test(job_ptr->details->req_node_bitmap, i)) {
				error("select_vipu Node %s is required but cannot be used for %u",
					  node_record_table_ptr[i]->name, job_ptr->job_id);
				return SLURM_ERROR;
			}
		}
	}
	return SLURM_SUCCESS;
}

/*
 * Determine how many nodes are available for a particular vipu gres
 *
 * Return number of available nodes
 *
 * IN job_ptr - pointer to job being considered for initiation,
 *                  set's start_time when job expected to start
 * IN vipu_plugin_id - vipu plugin id
 * IN vipu_plugin_model -- vipu plugin model
 * IN/OUT bitmap - usable nodes are set on input, nodes not required to
 *	satisfy the request are cleared, other left set
 * IN min_nodes - minimum count of nodes
 * IN req_nodes - requested (or desired) count of nodes
 * IN max_nodes - maximum count of nodes
 * IN mode - SELECT_MODE_RUN_NOW: try to schedule job now
 *           SELECT_MODE_TEST_ONLY: test if job can ever run
 *           SELECT_MODE_WILL_RUN: determine when and where job can run
 * IN preemptee_candidates - List of pointers to jobs which can be preempted.
 * IN preemptee_job_list - Pointer to list of job pointers. These are the
 *		jobs to be preempted to initiate the pending job. Not set
 *		if mode=SELECT_MODE_TEST_ONLY or input pointer is NULL.
 * IN exc_cores - bitmap of cores being reserved.
 */
static int _count_nodes_gres(job_record_t *job_ptr, uint32_t vipu_plugin_id,
				 const char *vipu_plugin_model, bitstr_t *node_bitmap,
				 uint32_t min_nodes, uint32_t max_nodes,
				 uint32_t req_nodes, uint16_t mode,
				 List preemptee_candidates,
				 List *preemptee_job_list, bitstr_t *exc_cores)
{
	const char *model = vipu_plugin_model;
	node_record_t *node_record;
	List gres_list;
	int i, i_first, i_last;
	int count = 0;
	uint32_t node_state;

	sv_debug("for job %u with model file %s",
		 job_ptr->job_id, vipu_plugin_model);

	sv_debug("for job %u with model %s",
		 job_ptr->job_id, model);

	i_first = bit_ffs(node_bitmap);
	if (i_first == -1)
		i_last = -2;
	else
		i_last = bit_fls(node_bitmap);

	for (i = i_first; i <= i_last; i++) {
		if (!bit_test(node_bitmap, i))
			continue;

		node_record = node_record_table_ptr[i];
		gres_list = node_record->gres_list;
		node_state = node_record->node_state;
		if (gres_plugin_node_test_model(vipu_plugin_id, (char *)model,
						gres_list) == SLURM_SUCCESS) {
			if (!(node_state & NODE_STATE_DRAIN) &&
				!(node_state & NODE_STATE_FAIL) &&
				!((node_state & NODE_STATE_BASE) == NODE_STATE_DOWN) &&
				!(node_state & NODE_STATE_MAINT)) {
				count++;
			} else {
				sv_debug("node is currently not available for job %u with model %s",
					 job_ptr->job_id, model);
			}
		}
	}
	return count;
}

/*
 * init() is called when the plugin is loaded, before any other functions
 * are called.
 */
extern int init(void)
{
	other_select_type_param = slurmctl_config.select_type_param;

	if (other_select_type_param & CR_OTHER_CONS_RES)
		plugin_id = SELECT_PLUGIN_VIPU_CONS_RES;
	else if (other_select_type_param & CR_OTHER_CONS_TRES)
		plugin_id = SELECT_PLUGIN_VIPU_CONS_TRES;
	else
		plugin_id = SELECT_PLUGIN_VIPU_LINEAR;

	slurm_mutex_lock(&vipu_config_list_lock);
	vipu_config_list = vipu_p_get_configs();
	if (!vipu_config_list || list_count(vipu_config_list) < 1) {
		error("select_vipu ERROR: no VIPU configuration file found");
		return SLURM_ERROR;
	}
	slurm_mutex_unlock(&vipu_config_list_lock);

	verbose("%s loaded", plugin_name);

	return SLURM_SUCCESS;
}

extern int fini(void)
{
	slurm_mutex_lock(&vipu_config_list_lock);
	if (vipu_config_list) {
		list_destroy(vipu_config_list);
		vipu_config_list = NULL;
	}
	slurm_mutex_unlock(&vipu_config_list_lock);
	pthread_mutex_destroy(&vipu_config_list_lock);
	return SLURM_SUCCESS;
}

/*
 * select_p_state_save saves any global node selection state
 */
extern int select_p_state_save(char *dir_name)
{
	sv_debug("called");
	return other_state_save(dir_name);
}

/*
 * select_p_state_restore restores any global node selection state
 */
extern int select_p_state_restore(char *dir_name)
{
	sv_debug("called");
	return other_state_restore(dir_name);
}

/*
 * select_p_job_init used at slurmctld daemon startup to synchronize plugin
 * (and node) state with that of currently active jobs.
 */
extern int select_p_job_init(List job_list)
{
	sv_debug("called");
	return other_job_init(job_list);
}

/*
 * select_p_node_ranking - generates node ranking for the nodes
 */
extern bool select_p_node_ranking(node_record_t *node_ptr, int node_cnt)
{
	sv_debug("called");
	return false;
}

/*
 * select_p_node_init - initializes node record data structure.
 * This function is called by the slurmctld daemon when the node records are
 * initially established and again when any nodes are added to or removed from
 * the data structure.
 */
extern int select_p_node_init()
{
	node_record_t *node_rec;
	int i;

	sv_debug("called");
	slurm_mutex_lock(&vipu_config_list_lock);

	for (i = 0; (node_rec = next_node(&i)); i++) {
		if (!node_rec->select_nodeinfo)
			node_rec->select_nodeinfo = select_g_select_nodeinfo_alloc();
	}

	slurm_mutex_unlock(&vipu_config_list_lock);
	return other_node_init();
}

/*
 * select_p_job_test - Given a specification of scheduling requirements,
 *	identify the nodes which "best" satisfy the request.
 *	"best" is defined as either single set of consecutive nodes satisfying
 *	the request and leaving the minimum number of unused nodes OR
 *	the fewest number of consecutive node sets
 * IN/OUT job_ptr - pointer to job being considered for initiation,
 *                  set's start_time when job expected to start
 * IN/OUT bitmap - usable nodes are set on input, nodes not required to
 *	satisfy the request are cleared, other left set
 * IN min_nodes - minimum count of nodes
 * IN req_nodes - requested (or desired) count of nodes
 * IN max_nodes - maximum count of nodes
 * IN mode - SELECT_MODE_RUN_NOW: try to schedule job now
 *           SELECT_MODE_TEST_ONLY: test if job can ever run
 *           SELECT_MODE_WILL_RUN: determine when and where job can run
 * IN preemptee_candidates - List of pointers to jobs which can be preempted.
 * IN/OUT preemptee_job_list - Pointer to list of job pointers. These are the
 *		jobs to be preempted to initiate the pending job. Not set
 *		if mode=SELECT_MODE_TEST_ONLY or input pointer is NULL.
 * IN exc_core_bitmap - bitmap of cores being reserved.
 * RET zero on success, EINVAL otherwise
 * globals (passed via select_p_node_init):
 *	node_record_count - count of nodes configured
 *	node_record_table_ptr - pointer to global node table
 * NOTE: the job information that is considered for scheduling includes:
 *	req_node_bitmap: bitmap of specific nodes required by the job
 *	contiguous: allocated nodes must be sequentially located
 *	num_cpus: minimum number of processors required by the job
 * NOTE: bitmap must be a superset of the job's required at the time that
 *	select_p_job_test is called
 */
extern int select_p_job_test(job_record_t *job_ptr, bitstr_t *bitmap,
				 uint32_t min_nodes, uint32_t max_nodes,
				 uint32_t req_nodes, uint16_t mode,
				 List preemptee_candidates,
				 List *preemptee_job_list,
				 bitstr_t *exc_core_bitmap)
{
	select_jobinfo_t *jobinfo;
	int retval = SLURM_SUCCESS;
	int count;
	uint64_t num_ipus, num_replicas;

	jobinfo = jobinfo_from_job_record(job_ptr);

	sv_debug("called");
	if (!jobinfo->ipus)
		num_ipus = get_count_for_gres(job_ptr->gres_list_req, "vipu");
	else
		num_ipus = jobinfo->ipus;

	if (num_ipus <= 0)
		return other_job_test(job_ptr, bitmap, min_nodes, max_nodes, req_nodes,
				  mode, preemptee_candidates, preemptee_job_list,
				  exc_core_bitmap);

	/*
	 * Set up job info and send to other plugin
	 */
	if (jobinfo->state == INIT) {
		// If the job has to be validated, just validate it and return
		num_replicas = get_count_for_gres(job_ptr->gres_list_req, "vipu_replica");
		jobinfo->num_replicas = num_replicas;
		jobinfo->ipus = num_ipus;
		info("%s: Requesting %lu ipus and %lu replicas",
			 plugin_type, jobinfo->ipus, jobinfo->num_replicas);
		jobinfo->state = VALIDATED;
		info("%s: IPU %pJ is detected and validated", plugin_type, job_ptr);
	/*
	 * Create partition
	 */
	}
	if (jobinfo->state == VALIDATED) {
		jobinfo->config = get_job_vipu_gres_model(job_ptr->gres_list_req);
		if (!jobinfo->config) {
			ListIterator config_iter;
			const char *config;

			slurm_mutex_lock(&vipu_config_list_lock);
			if (!vipu_config_list) {
				error("%s: No V-IPU configuration files available", plugin_type);
				slurm_mutex_unlock(&vipu_config_list_lock);
				return SLURM_ERROR;
			}
			config_iter = list_iterator_create(vipu_config_list);
			/* 
				* This is an IPU job for the default vipu model, try iterating over
				* the available vipu config files for creating partition for the job
				*/
			while ((config = (char *)list_next(config_iter))) {
				sv_debug("processing %s", config);
				// TODO: check it for nodes busy
				count = _count_nodes_gres(
					job_ptr, plugin_id, config, bitmap, min_nodes,
					max_nodes, req_nodes, mode, preemptee_candidates,
					preemptee_job_list, exc_core_bitmap);
				sv_debug("total available nodes for %s are %d",
						config, count);
				if (count > 0) {
					jobinfo->config = xstrdup(config);
					break;
				}
			}
			list_iterator_destroy(config_iter);
			slurm_mutex_unlock(&vipu_config_list_lock);
		}
		info("%s: %pJ creating partition config = %s",
			 plugin_type, job_ptr, jobinfo->config);
		if ((retval = request_job_resources(job_ptr))) {
			if (validate_ipu_job_for_config(jobinfo->config, jobinfo->ipus)) {
				error("%s: %pJ job configuration invalid", plugin_type, job_ptr);
				return SLURM_ERROR;
			}
			info("%s: %pJ partition not available; returning valid job to queue",
				plugin_type, job_ptr);
			return SLURM_SUCCESS;
		}
		jobinfo->state = PARTITION_CREATED;
	}
	/* 
	 * Final state checked every time job is returned to the queue
	 */
	if (jobinfo->state == PARTITION_CREATED) {
		char *model;
		info("%s: %pJ updating GRES model to %s",
			 plugin_type, job_ptr, jobinfo->config);
		model = get_job_vipu_gres_model(job_ptr->gres_list_req);
		if (!model) {
			info("%s: Setting gres model %s for %pJ", plugin_type, model, job_ptr);
			set_job_vipu_gres_model(job_ptr->gres_list_req, jobinfo->config);
		} else {
			xfree(model);
		}
		retval = _filter_nodes_gres(job_ptr, bitmap, min_nodes, max_nodes,
			req_nodes, mode, preemptee_candidates,
			preemptee_job_list, exc_core_bitmap);
		if (retval != SLURM_SUCCESS)
			return retval;
	}

	return other_job_test(job_ptr, bitmap, min_nodes, max_nodes, req_nodes,
				mode, preemptee_candidates, preemptee_job_list,
				exc_core_bitmap);
}

/*
 * select_p_job_begin - initiation of the specified job is about to begin.
 * This function is called immediately after select_p_job_test() successfully
 * completes for this job.
 */
extern int select_p_job_begin(job_record_t *job_ptr)
{
	select_jobinfo_t *jobinfo;
	jobinfo = jobinfo_from_job_record(job_ptr);

	sv_debug("called ipus = %lu model = %s partitions = %lu",
		 jobinfo->ipus, jobinfo->config, jobinfo->num_replicas);
	if (!jobinfo->ipus)
		return other_job_begin(job_ptr);

	if (jobinfo->state != PARTITION_CREATED) {
		error("select_vipu no partition exists for %u",
			job_ptr->job_id);
		return SLURM_ERROR;
	}

	/* Make sure that the partition is ACTIVE for the job */
	if (is_partition_active(job_ptr) != SLURM_SUCCESS) {
		error("select_vipu partition is still not in ACTIVE state for %u",
				job_ptr->job_id);
		return SLURM_ERROR;
	}
	// TODO: Sleep ?!
	sleep(1);
	if (copy_ipuof_config(job_ptr) != SLURM_SUCCESS) {
		error("select_vipu copying IPUoF config for %u failed",
				job_ptr->job_id);
		//TODO: Why we do not mutex here
		return SLURM_ERROR;
	}

	sv_debug("copying IPUoF copying IPUoF config for %u succeeded",
			 job_ptr->job_id);

	return other_job_begin(job_ptr);
}

/*
 * select_p_job_ready - Tests if resources are configured and ready for job
 * execution
 */
extern int select_p_job_ready(job_record_t *job_ptr)
{
	return other_job_ready(job_ptr);
}

/*
 * select_p_job_resized - Removes the specified node from the job's allocation
 */
extern int select_p_job_resized(job_record_t *job_ptr,
				node_record_t *node_ptr)
{
	return other_job_resized(job_ptr, node_ptr);
}

/*
 * select_p_job_expand - Reports the ability of this select plugin to expand
 * jobs.
 */
extern int select_p_job_expand(job_record_t *from_job_ptr,
	job_record_t *to_job_ptr)
{
	return other_job_expand(from_job_ptr, to_job_ptr);
}

/*
 * select_p_job_signal - Signals the specified job
 */
extern int select_p_job_signal(job_record_t *job_ptr, int signal)
{
	xassert(job_ptr);
	sv_debug(" job_signal %d received for job %u", signal, job_ptr->job_id);
	switch (signal) {
	case SIGKILL:
		// check if there is a partition created for the job and delete it
		if (release_job_resources(job_ptr) != SLURM_SUCCESS) {
			error("select_vipu could not release job resources for the job %u on signal %d",
				  job_ptr->job_id, signal);
		}
		break;
	}
	return other_job_signal(job_ptr, signal);
}

/*
 * select_p_job_mem_confirm - Confirms that a job's memory allocation is still
 * valid after a node is restarted.
extern int select_p_job_mem_confirm(job_record_t *job_ptr)
{
	xassert(job_ptr);
	return other_job_mem_confirm(job_ptr);
}
*/

/*
 * select_p_job_fini - Notes the termination of the specified job.
 * This function is called as the termination process for the job begins (prior
 * to killing the tasks).
 */
extern int select_p_job_fini(job_record_t *job_ptr)
{
	xassert(job_ptr);
	xassert(job_ptr->select_jobinfo);
	xassert(job_ptr->select_jobinfo->data);

	if (release_job_resources(job_ptr) != SLURM_SUCCESS)
		return SLURM_ERROR;
	other_job_fini(job_ptr);
	return SLURM_SUCCESS;
}

/*
 * select_p_job_suspend - Notes the Suspension of a job
 */
extern int select_p_job_suspend(job_record_t *job_ptr, bool indf_susp)
{
	sv_debug("%u", job_ptr->job_id);
	return other_job_suspend(job_ptr, indf_susp);
}

/*
 * other_job_resume - Notes the resumption of a job
 */
extern int select_p_job_resume(job_record_t *job_ptr, bool indf_susp)
{
	sv_debug("%u", job_ptr->job_id);
	return other_job_resume(job_ptr, indf_susp);
}

/*
 * select_p_step_pick_nodes -  If the select plugin needs to select nodes for a
 * job step, then do so here.
 */
extern bitstr_t *select_p_step_pick_nodes(job_record_t *job_ptr,
					  select_jobinfo_t *step_jobinfo,
					  uint32_t node_count,
					  bitstr_t **avail_nodes)
{
	sv_debug("%u", job_ptr->job_id);
	return other_step_pick_nodes(job_ptr, step_jobinfo, node_count, avail_nodes);
}

/*
 * select_p_step_start - Start the job step
 */
extern int select_p_step_start(step_record_t *step_ptr)
{
	sv_debug(" Starting step for the job: %u, step: %u",
		step_ptr->job_ptr->job_id, step_ptr->step_id.step_id);
	return other_step_start(step_ptr);
}

/*
 * select_p_step_finish - Notes that a job step is completing execution
 */
extern int select_p_step_finish(step_record_t *step_ptr, bool killing_step)
{
	sv_debug("%u", step_ptr->job_ptr->job_id);
	//TODO: See neat comment in select_cray_aries.c:1733. Not sure it is needed here.
	if (killing_step)
		jobacct_storage_g_step_complete(acct_db_conn, step_ptr);

	return other_step_finish(step_ptr, killing_step);
}

/*
 * select_p_select_nodeinfo_alloc - Allocates a buffer for select plugin
 * specific information about a node
 */
extern select_nodeinfo_t *select_p_select_nodeinfo_alloc(void)
{
	select_nodeinfo_t *nodeinfo = xmalloc(sizeof(struct select_nodeinfo));

	sv_debug("entry");
	nodeinfo->magic = NODEINFO_MAGIC;
	nodeinfo->other_nodeinfo = other_select_nodeinfo_alloc();

	return nodeinfo;
}

/*
 * select_p_select_nodeinfo_free - Frees buffer for select plugin specific
 * information about a node
 */
extern int select_p_select_nodeinfo_free(select_nodeinfo_t *nodeinfo)
{
	sv_debug("entry");
	if (nodeinfo) {
		other_select_nodeinfo_free(nodeinfo->other_nodeinfo);
		xfree(nodeinfo);
	}
	return SLURM_SUCCESS;
}

/*
 * select_p_select_nodeinfo_pack - Packs select plugin specific information
 * about a node into a buffer for node queries.
 */
extern int select_p_select_nodeinfo_pack(select_nodeinfo_t *nodeinfo,
					 buf_t *buffer,
					 uint16_t protocol_version)
{
	int rc = SLURM_ERROR;

	sv_debug("entry");
	if (!nodeinfo) {
		error("select_vipu %s: nodeinfo is NULL", __func__);
		rc = other_select_nodeinfo_pack(NULL, buffer, protocol_version);
	} else {
		rc = other_select_nodeinfo_pack(nodeinfo->other_nodeinfo,
						buffer,	protocol_version);
	}
	return rc;
}

/*
 * select_p_select_nodeinfo_unpack - Unpacks select plugin specific information
 * about a node from a buffer.
 */
extern int select_p_select_nodeinfo_unpack(select_nodeinfo_t **nodeinfo_pptr,
					   buf_t *buffer,
					   uint16_t protocol_version)
{
	int rc = SLURM_ERROR;
	select_nodeinfo_t *nodeinfo = xmalloc(sizeof(struct select_nodeinfo));

	sv_debug("entry");
	nodeinfo->magic = NODEINFO_MAGIC;
	rc = other_select_nodeinfo_unpack(&nodeinfo->other_nodeinfo, buffer,
					  protocol_version);
	if (rc != SLURM_SUCCESS) {
		select_p_select_nodeinfo_free(nodeinfo);
		//TODO: Do we need to clear nodeinfo_ptr?
		return SLURM_ERROR;
	}
	*nodeinfo_pptr = nodeinfo;
	return SLURM_SUCCESS;
}

/*
 * select_p_select_nodeinfo_set_all - Updates select plugin specific information
 * about every node.
 */
extern int select_p_select_nodeinfo_set_all(void)
{
	int i;

	sv_debug("entry");
	for (i = 0; i < node_record_count; i++) {
		node_record_t *node_ptr = node_record_table_ptr[i];

		node_ptr->node_state &= (~NODE_STATE_NET);
	}

	return other_select_nodeinfo_set_all();
}

/*
 * select_p_select_nodeinfo_set - Resets select plugin specific information
 * about a job. Called by slurmctld daemon after that job's state has been
 * restored (at startup) or job has been scheduled.
 */
extern int select_p_select_nodeinfo_set(job_record_t *job_ptr)
{
	sv_debug("entry");
	return other_select_nodeinfo_set(job_ptr);
}

/*
 * select_p_select_nodeinfo_get - Gets information from a select plugin's node
 * specific data structure.
 */
extern int select_p_select_nodeinfo_get(select_nodeinfo_t *nodeinfo,
					enum select_nodedata_type dinfo,
					enum node_states state, void *data)
{
	int rc = SLURM_SUCCESS;
	select_nodeinfo_t **select_nodeinfo = (select_nodeinfo_t **)data;

	sv_debug("entry");
	if (!nodeinfo) {
		error("select/vipu nodeinfo_get: nodeinfo not set");
		return SLURM_ERROR;
	}
	if (nodeinfo->magic != NODEINFO_MAGIC) {
		error("select/vipu nodeinfo_get: nodeinfo magic bad");
		return SLURM_ERROR;
	}

	switch (dinfo) {
	case SELECT_NODEDATA_PTR:
		*select_nodeinfo = nodeinfo->other_nodeinfo;
		break;
	default:
		rc = other_select_nodeinfo_get(nodeinfo->other_nodeinfo,
						   dinfo, state, data);
		break;
	}
	return rc;
}

/*
 * select_p_select_jobinfo_alloc - Allocates a buffer for select plugin specific
 * information about a job.
 */
extern select_jobinfo_t *select_p_select_jobinfo_alloc(void)
{
	select_jobinfo_t *jobinfo = xcalloc(1, sizeof(struct select_jobinfo));

	if (!jobinfo)
		return NULL;
	sv_debug("entry");
	jobinfo->magic = JOBINFO_MAGIC;
	jobinfo->other_jobinfo = other_select_jobinfo_alloc();

	return jobinfo;
}

/*
 * select_p_select_jobinfo_free - Frees buffer for select plugin specific
 * information about a job.
 */
extern int select_p_select_jobinfo_free(select_jobinfo_t *jobinfo)
{
	int rc = SLURM_SUCCESS;

	sv_debug("entry");
	if (jobinfo) {
		if (jobinfo->magic != JOBINFO_MAGIC) {
			error("select/vipu jobinfo_free: jobinfo magic bad");
			//TODO: No SLURM ret code?
			return SLURM_ERROR;
		}

		jobinfo->magic = 0;
		jobinfo->state = INIT;
		xfree(jobinfo->config);
		jobinfo->config = NULL;
		rc = other_select_jobinfo_free(jobinfo->other_jobinfo);
		xfree(jobinfo);
	}

	return rc;
}

/*
 * select_p_select_jobinfo_set - Sets a field in the select plugin specific
 * information about a job.
 */
extern int select_p_select_jobinfo_set(select_jobinfo_t *jobinfo,
					   enum select_jobdata_type data_type,
					   void *data)
{
	sv_debug("entry");
	if (!jobinfo) {
		error("select/vipu jobinfo_set: jobinfo not set");
		return SLURM_ERROR;
	}
	if (jobinfo->magic != JOBINFO_MAGIC) {
		error("select/vipu jobinfo_set: jobinfo magic bad");
		return SLURM_ERROR;
	}

	switch (data_type) {
	case SELECT_JOBDATA_CLEANING:
		break;
	case SELECT_JOBDATA_RELEASED:
		break;
	case SELECT_JOBDATA_NETWORK:
		break;
	default:
		return other_select_jobinfo_set(jobinfo,
						data_type, data);
	}

	return SLURM_SUCCESS;
}

/*
 * select_p_select_jobinfo_get - Gets a field from the select plugin specific
 * information about a job.
 */
extern int select_p_select_jobinfo_get(select_jobinfo_t *jobinfo,
				       enum select_jobdata_type data_type,
				       void *data)
{
	sv_debug("entry");
	/* not implemented */
	return SLURM_ERROR;
}

/*
 * select_p_select_jobinfo_copy - Copies the buffer containing select plugin
 * specific information about a job.
 */
extern select_jobinfo_t *
select_p_select_jobinfo_copy(select_jobinfo_t *jobinfo)
{
	struct select_jobinfo *rc = NULL;

	sv_debug("entry");
	if (!jobinfo)
		return NULL;

	if (jobinfo->magic != JOBINFO_MAGIC) {
		error("select/vipu jobinfo_copy: jobinfo magic bad");
		return NULL;
	}
	rc = xcalloc(1, sizeof(struct select_jobinfo));
	if (!rc)
		return rc;
	rc->magic = JOBINFO_MAGIC;
	rc->ipus = jobinfo->ipus;
	rc->num_replicas = jobinfo->num_replicas;
	rc->state = jobinfo->state;
	rc->config = xstrdup(jobinfo->config);
	return rc;
}

/*
 * select_p_select_jobinfo_pack - Packs into a buffer the contents of the select
 * plugin specific information about a job.
 */
extern int select_p_select_jobinfo_pack(select_jobinfo_t *jobinfo, buf_t *buffer,
					uint16_t protocol_version)
{
	sv_debug("entry");
	if (jobinfo) {
		pack64(jobinfo->ipus, buffer);
		pack64(jobinfo->num_replicas, buffer);
		pack32(jobinfo->state, buffer);
		if (jobinfo->state == PARTITION_CREATED)
			packstr(jobinfo->config, buffer);

		return other_select_jobinfo_pack(jobinfo->other_jobinfo, buffer,
						 protocol_version);
	}
	pack64(0, buffer);
	pack64(0, buffer);
	pack32(INIT, buffer);
	// TODO: may be should pack some data
	return other_select_jobinfo_pack(NULL, buffer, protocol_version);
}

/*
 * select_p_select_jobinfo_unpack - Unpacks from a buffer the contents of the
 * select plugin specific information about a job.
 */
extern int select_p_select_jobinfo_unpack(select_jobinfo_t **jobinfo_pptr,
					  buf_t *buffer, uint16_t protocol_version)
{
	int rc;
	select_jobinfo_t *jobinfo = xcalloc(1, sizeof(struct select_jobinfo));

	sv_debug("entry");
	if (!jobinfo)
		return SLURM_ERROR;

	jobinfo->magic = JOBINFO_MAGIC;

	rc = other_select_jobinfo_unpack(&jobinfo->other_jobinfo, buffer,
					 protocol_version);
	if (rc != SLURM_SUCCESS) {
		xfree(jobinfo);
		return SLURM_ERROR;
	}

	unpack64(&jobinfo->ipus, buffer);
	unpack64(&jobinfo->num_replicas, buffer);
	unpack32(&jobinfo->state, buffer);
	if (jobinfo->state == PARTITION_CREATED) {
		uint32_t len;

		unpackstr_xmalloc(&jobinfo->config, &len, buffer);
		jobinfo->config[len] = 0;
		sv_debug("--------------------- %s ------------------- (%u)", jobinfo->config, len);
	}
	sv_debug("unpack %lu %lu %i",
		 jobinfo->ipus, jobinfo->num_replicas, jobinfo->state);

	*jobinfo_pptr = jobinfo;
	return SLURM_SUCCESS;
}

/*
 * select_p_select_jobinfo_sprint - Prints the contents of the select plugin
 * specific information about a job.
 */
extern char *select_p_select_jobinfo_sprint(select_jobinfo_t *jobinfo,
						char *buf, size_t size, int mode)
{
	sv_debug("entry");
	if (!buf || !size)
		return NULL;
	snprintf(buf, size, "VIPU: v:%d i:%lu r:%lu",
	 jobinfo->state, jobinfo->ipus, jobinfo->num_replicas);
	buf[0] = '\0';
	return buf;
}

/*
 * select_p_select_jobinfo_xstrdup - Print the contents of the select plugin
 * specific information about a job. The return value must be released using the
 * xfree() function.
 */
extern char *select_p_select_jobinfo_xstrdup(select_jobinfo_t *jobinfo,
						 int mode)
{
	char *buf = NULL;

	sv_debug("entry");
	if ((mode != SELECT_PRINT_DATA) && jobinfo &&
		(jobinfo->magic != JOBINFO_MAGIC)) {
		error("select/vipu jobinfo_xstrdup: jobinfo magic bad");
		return NULL;
	}

	if (!jobinfo) {
		if (mode != SELECT_PRINT_HEAD) {
			error("select/vipu jobinfo_xstrdup: jobinfo bad");
			return NULL;
		}
		return NULL;
	}

	xstrcat(buf, other_select_jobinfo_xstrdup(jobinfo->other_jobinfo, mode));

	return buf;
}

/*
 * select_p_get_info_from_plugin - Gets plugin-specific information about a job.
 */
extern int select_p_get_info_from_plugin(enum select_plugindata_info dinfo,
					 job_record_t *job_ptr, void *data)
{
	sv_debug("entry");	
	return other_get_info_from_plugin(dinfo, job_ptr, data);
}

/*
 * select_p_update_node_config - Notes that a node has registered with a
 * different configuration than previously registered. For example, the node was
 * configured with 1GB of memory in slurm.conf, but actually registered with 2GB
 * of memory.
 */
extern int select_p_update_node_config(int index)
{
	sv_debug("entry");
	return other_update_node_config(index);
}

/*
 * select_p_reconfigure - Used to notify plugin of change in partition
 * configuration or general configuration change. The plugin will test global
 * variables for changes as appropriate.
 */
extern int select_p_reconfigure(void)
{
	sv_debug("entry");
	return other_reconfigure();
}

/*
 * select_p_resv_test - Identifies the node which best satisfy a reservation
 * request taking system topology into consideration if applicable.
 */
extern bitstr_t *select_p_resv_test(resv_desc_msg_t *resv_desc_ptr,
					uint32_t node_cnt, bitstr_t *avail_bitmap,
					bitstr_t **core_bitmap)
{
	sv_debug("entry");
	return other_resv_test(resv_desc_ptr, node_cnt, avail_bitmap, core_bitmap);
}


